"""
RasgoQL main class
"""
from typing import List, Type

import pandas as pd
import rasgotransforms as rtx

from rasgoql.data import DataWarehouse
from rasgoql.primitives import Dataset, SQLChain, TransformTemplate
from rasgoql.utils.messaging import set_verbose
from rasgoql.version import __version__

class RasgoQL:
    """
    Entry class for rasgoql
    """
    __version__ = __version__

    def __init__(
            self,
            connection: Type[DataWarehouse],
            credentials: dict
        ):
        self.credentials = credentials
        self._dw = connection()
        self._dw.connect(credentials)

    def disconnect_dw(self):
        """
        Closes a connection to the Data Warehouse
        """
        self._dw.close_connection()

    def list_tables(
            self,
            database: str = None,
            schema: str = None
        ):
        """
        Returns a list of tables in your Data Warehouse
        """
        return self._dw.list_tables(database, schema)

    def list_transforms(self) -> List[TransformTemplate]:
        """
        Returns a list of RasgoQL transforms
        """
        return rtx.serve_rasgo_transform_templates(self._dw.dw_type)

    def dataset(
            self,
            fqtn: str
        ) -> Dataset:
        """
        Returns a Dataset connected to the Cloud Data Warehouse
        """
        return Dataset(fqtn, self._dw)

    def dataset_from_df(
            self,
            df: pd.DataFrame,
            table_name: str,
            method: str = None
        ) -> Dataset:
        """
        Writes a pandas Dataframe to a table in your Data Warehouse
        and returns it as a Dataset

        Params:
        `df`: pandas DataFrame:
            DataFrame to upload
        `table_name`: str:
            Name for the new table
        `method`: str
            Values: [append, replace]
            when this table already exists in your DataWarehouse,
            pass append: to add dataframe rows to it
            pass replace: to overwrite it with dataframe rows
                WARNING: This will completely overwrite data in the existing table
        """
        fqtn = self._dw.save_df(df, table_name, method)
        return Dataset(fqtn, self._dw)

    def define_transform(
            self,
            name: str
        ) -> str:
        """
        Returns full details of a RasgoQL transform
        """
        udt: TransformTemplate = None
        for t in rtx.serve_rasgo_transform_templates(self._dw.dw_type):
            if t.name == name:
                udt = t
        if udt:
            return udt.define()
        raise ValueError(f'{name} is not a valid Tranform name. ' \
                           'Run `.list_transforms()` to see available transforms.')

    def query(
            self,
            sql: str
        ):
        """
        Execute a query against your Data Warehouse
        """
        return self._dw.execute_query(sql)

    def query_into_df(
            self,
            sql: str
        ) -> pd.DataFrame:
        """
        Execute a query against your Data Warehouse and return results in a pandas DataFrame
        """
        return self._dw.execute_query(sql, response='df')

    def set_verbose(
            self,
            value: bool
        ) -> None:
        """
        Turn verbose logging on or off

        value: bool
            True = log more info about SQL and primitive activities
            False = log almost nothing
        """
        set_verbose(value)

    def sqlchain(
            self,
            fqtn: str,
            namespace: str = None
        ) -> SQLChain:
        """
        Returns a SQLChain connected to the Cloud Data Warehouse
        """
        return SQLChain(
            entry_table=Dataset(
                fqtn,
                self._dw
            ),
            namespace=namespace or self._dw.default_namespace,
            dw=self._dw
        )



"""
Base DataWarehouse classes
"""
from abc import ABC
from typing import Union

import pandas as pd


class DWCredentials(ABC):
    """
    Base DW Credentials
    """
    dw_type = None

    @classmethod
    def from_dict(
            cls,
            source_dict: dict
        ) -> 'DWCredentials':
        """
        Creates an instance of this Class from a dict
        """
        raise NotImplementedError()

    @classmethod
    def from_env(
            cls,
            filepath: str = None
        ) -> 'DWCredentials':
        """
        Creates an instance of this Class from a .env file on your machine
        """
        raise NotImplementedError()

    def to_dict(self) -> dict:
        """
        Returns a dict of the credentials
        """
        raise NotImplementedError()

    def to_env(
            self,
            filepath: str,
            overwrite: bool
        ):
        """
        Saves credentials to a .env file on your machine
        """
        raise NotImplementedError()

class DataWarehouse(ABC):
    """
    Base DW Class
    """
    dw_type = None
    credentials_class = DWCredentials

    def __init__(self):
        self.credentials = None
        self.connection = None

    def change_namespace(
            self,
            namespace: str
    ):
        """
        Change the default namespace of this connection
        """
        raise NotImplementedError()

    def connect(
            self,
            credentials: Union[dict, DWCredentials]
        ):
        """
        Connect to this DataWarehouse
        """
        raise NotImplementedError()

    def close_connection(self):
        """
        Close connection to this DataWarehouse
        """
        raise NotImplementedError()

    def create(
            self,
            sql: str,
            fqtn: str,
            table_type: str = 'VIEW',
            overwrite: bool = False
        ):
        """
        Create a view or table from given SQL
        """
        raise NotImplementedError()

    @property
    def default_namespace(self) -> str:
        """
        Returns the default namespace of this connection
        """
        raise NotImplementedError()

    def execute_query(
            self,
            sql: str,
            response: str = 'tuple',
            acknowledge_risk: bool = False
        ):
        """
        Run a query against this DataWarehouse
        """
        raise NotImplementedError()

    def get_ddl(
            self,
            fqtn: str
        ) -> str:
        """
        Returns the create statement for a table or view
        """
        raise NotImplementedError()

    def get_object_details(
            self,
            fqtn: str
        ) -> tuple:
        """
        Return details of a table or view in this DataWarehouse
        """
        raise NotImplementedError()

    def get_schema(
            self,
            fqtn: str,
            create_sql: str = None
        ) -> dict:
        """
        Return the schema of a table or view
        """
        raise NotImplementedError()

    def list_tables(
            self,
            database: str = None,
            schema: str = None
        ):
        """
        List all available tables in this DataWarehouse
        """
        raise NotImplementedError()

    def preview(
            self,
            sql: str,
            limit: int = 10
        ) -> pd.DataFrame:
        """
        Returns 10 records into a pandas DataFrame
        """
        raise NotImplementedError()

    def save_df(
            self,
            df: pd.DataFrame,
            fqtn: str,
            method: str = None
        ):
        """
        Creates a table in this DataWarehouse from a pandas Dataframe
        """
        raise NotImplementedError()

    # ---------------------------
    # Core Data Warehouse helpers
    # ---------------------------
    def _table_exists(
            self,
            fqtn: str
        ) -> bool:
        """
        Check for existence of fqtn in this Data Warehouse and return a boolean
        """
        raise NotImplementedError()

    def _validate_namespace(
            self,
            namespace: str
        ):
        """
        Checks a namespace string for compliance with this DataWarehouse format
        """
        raise NotImplementedError()



"""
Helpful DataFrame utilities
"""
import pandas as pd
from pandas.io.json import build_table_schema

from rasgoql.utils.sql import cleanse_sql_name


def build_dataframe_schema(
        df: pd.DataFrame,
        include_index=False
    ) -> dict:
    """
    Returns a dict representing the schema of a dataframe
    """
    schema_list = build_table_schema(df)
    if not include_index:
        return {column['name']: column
                for column in schema_list['fields'] if column['name'] != 'index'}
    return {column['name']: column
            for column in schema_list['fields']}

def cleanse_sql_dataframe(
        df: pd.DataFrame
    ):
    """
    Renames all columns in a pandas dataframe to SQL compliant names in place
    """
    df.rename(columns={r: cleanse_sql_name(r) for r in build_dataframe_schema(df)},
              inplace=True)

def generate_dataframe_ddl(
        df: pd.DataFrame,
        table_name: str
    ) -> str:
    """
    Generates a SQL statement to create a table matching the schema of a dataframe
    """
    create_statement = "CREATE OR REPLACE TABLE"
    return pd.io.sql.get_schema(df, table_name) \
        .replace("CREATE TABLE", create_statement) \
        .replace('"', '')

def map_pandas_data_type(
        data_type: str
    ) -> str:
    """
    For a given pandas data type, return the equivalent python type
    """
    if data_type.startswith('float'):
        return 'float'
    if data_type.startswith('datetime'):
        return 'datetime'
    if data_type.startswith('int'):
        return 'integer'
    return data_type

def tag_dataframe(
        df: pd.DataFrame,
        attribute: dict
    ):
    """
    Tags a dataframe with a user-defined attribute
    """
    write_back_attrs = df.attrs or {}
    write_back_attrs.update(attribute)
    df.attrs = write_back_attrs



"""
Enums
"""
from enum import Enum

from rasgoql.errors import ParameterException


def _wrap_in_quotes(string: str) -> str:
    return f"'{string}'"

class DWType(Enum):
    """
    Supported Data Warehouses
    """
    BIGQUERY = 'bigquery'
    #FUTURE: POSTGRES = 'postgres'
    SNOWFLAKE = 'snowflake'

def check_data_warehouse(input_value: str):
    """
    Warn if an incorrect Data Warehouse is passed
    """
    supported_dws = _wrap_in_quotes("', '".join([e.value for e in DWType]))
    try:
        DWType[input_value.upper()]
    except Exception:
        raise ParameterException(f'data_warehouse parameter accepts values: {supported_dws}')
    return input_value.upper()


class TableState(Enum):
    """
    State of a table
    """
    IN_DW = 'in dw'
    IN_MEMORY = 'in memory'
    UNKNOWN = 'unknown'

def check_table_state(input_value: str):
    """
    Warn if an incorrect table state is passed
    """
    table_states = _wrap_in_quotes("', '".join([e.value for e in TableState]))
    try:
        TableState[input_value.upper()]
    except Exception:
        raise ParameterException(f'table_state parameter accepts values: {table_states}')
    return input_value.upper()


class TableType(Enum):
    """
    Type of table in a DW
    """
    TABLE = 'table'
    VIEW = 'view'
    UNKNOWN = 'unknown'

def check_table_type(input_value: str):
    """
    Warn if an incorrect table type is passed
    """
    table_types = _wrap_in_quotes("', '".join([e.value for e in TableType]))
    try:
        TableType[input_value.upper()]
    except Exception:
        raise ParameterException(f'table_type parameter accepts values: {table_types}')
    return input_value.upper()


class RenderMethod(Enum):
    """
    Ways to render a sql chain
    """
    SELECT = 'select'
    TABLE = 'table'
    VIEW = 'view'
    VIEWS = 'views'

def check_render_method(input_value: str):
    """
    Warn if an incorrect render method is passed
    """
    render_methods = _wrap_in_quotes("', '".join([e.value for e in RenderMethod]))
    try:
        RenderMethod[input_value.upper()]
    except Exception:
        raise ParameterException(f'render_method parameter accepts values: {render_methods}')
    return input_value.upper()


class ResponseType(Enum):
    """
    Formats to return query results
    """
    DICT = 'dict'
    DF = 'df'
    TUPLE = 'tuple'
    NONE = 'none'

def check_response_type(input_value: str):
    """
    Warn if an incorrect response type is passed
    """
    response_types = _wrap_in_quotes("', '".join([e.value for e in ResponseType]))
    try:
        ResponseType[input_value.upper()]
    except Exception:
        raise ParameterException(f'response parameter accepts values: {response_types}')
    return input_value.upper()


class WriteMethod(Enum):
    """
    Ways to write data
    """
    APPEND = 'append'
    REPLACE = 'replace'
    UPSERT = 'upsert'

def check_write_method(input_value: str):
    """
    Warn if an incorrect write method is passed
    """
    write_methods = _wrap_in_quotes("', '".join([e.value for e in WriteMethod]))
    try:
        WriteMethod[input_value.upper()]
    except Exception:
        raise ParameterException(f'method parameter accepts values: {write_methods}')
    return input_value.upper()



"""
Helpful sql utilities
"""
import random
import re
import string


def cleanse_sql_data_type(
        dtype: str
    ) -> str:
    """
    Converts pandas data types to SQL compliant data type
    """
    if dtype.lower() in ["object", "text", "variant"]:
        return "string"
    else:
        return dtype.lower()

def cleanse_sql_name(
        name: str
) -> str:
    """
    Converts a string to a SQL compliant value
    """
    return name.replace(" ", "_").replace("-", "_").replace('"', '').replace(".", "_")

def is_scary_sql(
        sql: str
) -> bool:
    """
    Checks a SQL string for presence of injection keywords
    """
    if any(word in sql.upper() for word in SQL_INJECTION_KEYWORDS):
        return True
    return False

def is_restricted_sql(
        sql: str
) -> bool:
    """
    Checks a SQL string for presence of dangerous keywords
    """
    if any(word in sql.upper() for word in SQL_RESTRICTED_KEYWORDS):
        return True
    return False

def magic_fqtn_handler(
        possible_fqtn: str,
        default_namespace: str
    ) -> str:
    """
    Makes all of your wildest dreams come true... well not *that* one
    """
    input_db, input_schema, table = parse_fqtn(possible_fqtn, default_namespace, False)
    default_database, default_schema = parse_namespace(default_namespace)
    database = input_db or default_database
    schema = input_schema or default_schema
    return make_fqtn(database, schema, table)

def make_fqtn(
        database: str,
        schema: str,
        table: str
) -> str:
    """
    Accepts component parts and returns a fully qualified table string
    """
    return f"{database}.{schema}.{table}"

def make_namespace_from_fqtn(
        fqtn: str
) -> str:
    """
    Accepts component parts and returns a fully qualified namespace string
    """
    database, schema, _ = parse_fqtn(fqtn)
    return f"{database}.{schema}"

def parse_fqtn(
        fqtn: str,
        default_namespace: str = None,
        strict: bool = True
) -> tuple:
    """
    Accepts a possible fully qualified table string and returns its component parts
    """
    if strict:
        fqtn = validate_fqtn(fqtn)
        return (* fqtn.split("."),)
    database, schema = parse_namespace(default_namespace)
    if fqtn.count(".") == 2:
        return (* fqtn.split("."),)
    if fqtn.count(".") == 1:
        return (database, * fqtn.split("."),)
    if fqtn.count(".") == 0:
        return (database, schema, fqtn)
    raise ValueError(f'{fqtn} is not a well-formed fqtn')

def parse_namespace(
        namespace: str
) -> tuple:
    """
    Accepts a possible namespace string and returns its component parts
    """
    namespace = validate_namespace(namespace)
    return tuple(namespace.split("."))

def random_table_name() -> str:
    """
    Returns a random unique table name prefixed with "RQL"
    """
    return 'RQL_'+''.join(random.choice(string.ascii_uppercase) for x in range(10))

def validate_fqtn(
        fqtn: str
    ) -> str:
    """
    Accepts a possible fully qualified table string and decides whether it is well formed
    """
    if re.match(r'\w+\.\w+\.\w+', fqtn):
        return fqtn
    raise ValueError(f'{fqtn} is not a well-formed fqtn')

def validate_namespace(
        namespace: str
    ) -> bool:
    """
    Accepts a possible namespace string and decides whether it is well formed
    """
    if re.match(r'\w+\.\w+', namespace):
        return namespace
    raise ValueError(f'{namespace} is not a well-formed namespace')

def wrap_table(
        parent_table: str
    )-> str:
    """
    Calculates a unique table string
    """
    prefix = 'RQL'
    suffix = ''.join(random.choice(string.ascii_uppercase) for x in range(5))
    if parent_table.startswith('RQL'):
        return f"{parent_table}_{suffix}"
    return f"{prefix}_{parent_table}_{suffix}"


SQL_RESTRICTED_CHARACTERS = [
    ' ', '-', ';'
]

SQL_INJECTION_KEYWORDS = [
    'DELETE',
    'TRUNCATE',
    'DROP',
    'ALTER',
    'UPDATE',
    'INSERT',
    'MERGE'
]

SQL_RESTRICTED_KEYWORDS = [
    'ACCOUNT', 'ALL', 'ALTER', 'AND', 'ANY', 'AS', 'BETWEEN', 'BY',
    'CASE', 'CAST', 'CHECK', 'COLUMN', 'CONNECT', 'CONNECTION', 'CONSTRAINT',
    'CREATE', 'CROSS', 'CURRENT', 'CURRENT_DATE', 'CURRENT_TIME', 'CURRENT_TIMESTAMP',
    'CURRENT_USER', 'DATABASE', 'DELETE', 'DISTINCT', 'DROP', 'ELSE', 'EXISTS', 'FALSE',
    'FOLLOWING', 'FOR', 'FROM', 'FULL', 'GRANT', 'GROUP', 'GSCLUSTER', 'HAVING', 'ILIKE',
    'IN', 'INCREMENT', 'INNER', 'INSERT', 'INTERSECT', 'INTO', 'IS', 'ISSUE', 'JOIN', 'LATERAL',
    'LEFT', 'LIKE', 'LOCALTIME', 'LOCALTIMESTAMP', 'MINUS', 'NATURAL', 'NOT', 'NULL', 'OF', 'ON',
    'OR', 'ORDER', 'ORGANIZATION', 'QUALIFY', 'REGEXP', 'REVOKE', 'RIGHT', 'RLIKE', 'ROW', 'ROWS',
    'SAMPLE', 'SCHEMA', 'SELECT', 'SET', 'SOME', 'START', 'TABLE', 'TABLESAMPLE', 'THEN', 'TO', 'TRIGGER',
    'TRUE', 'TRY_CAST', 'UNION', 'UNIQUE', 'UPDATE', 'USING', 'VALUES', 'VIEW', 'WHEN', 'WHENEVER', 'WHERE', 'WITH'
]



"""
Manage Data Warehouse specific imports
"""
import logging
import pkg_resources
import requests

logging.basicConfig()
logger = logging.getLogger('RasgoQL Dependency Warning')
logger.setLevel(logging.INFO)

### ------------
# Note to users:
# This module is used to allow rasgoql to install the minimum dependencies
# needed to run your flavor of DataWarehouse

# We support downloading extas packages for each DW: e.g. `rasgoql[snowflake]`
# so your machine is not cluttered with unused google or sqlalchemy packages.
# In doing so, we run the risk of not having required dependencies at runtime

# This module attempts to import all possible datawarehouse packages, and
# fails gracefully if they are not present
# RasgoQL will use these empty aliases to raise a PackageDependencyWarning
# from the class or function that needs the import
### ------------

# Check for latest version of rasgotransforms package and prompt user
transforms_version = pkg_resources.get_distribution("rasgotransforms").version
try:
    response = requests.get('https://pypi.org/pypi/rasgotransforms/json')
    latest_version = response.json()['info']['version']
except:
    latest_version = transforms_version
if transforms_version != latest_version:
    logger.warning('You are not running the lastest version of rasgotransforms. '
                   'RasgoQL relies on this package to serve transform templates. '
                   'Please consider running `pip install rasgotransforms --upgrade` '
                   'to download our full library of SQL transforms.')

# Attempt safe imports of all DW packages so we can warn later if they are missing
try:
    from snowflake import connector as sf_connector
    from snowflake.connector.pandas_tools import write_pandas
except ImportError:
    sf_connector = None
    write_pandas = None

try:
    from google.cloud import bigquery as bq
    from google.cloud import exceptions as gcp_exc
    from google.oauth2 import service_account as gcp_svc
except ImportError:
    bq = None
    gcp_exc = None
    gcp_svc = None

try:
    from google_auth_oauthlib import flow as gcp_flow
except ImportError:
    gcp_flow = None



"""
pypi setup
"""
import os
from setuptools import setup

from rasgoql.version import __version__

_here = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(_here, 'DESCRIPTION.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(os.path.join(_here, 'requirements.txt'), encoding='utf-8') as f:
    req_lines = f.read()
    requirements = req_lines.splitlines()

with open(os.path.join(_here, 'requirements-snowflake.txt'), encoding='utf-8') as f:
    req_lines = f.read()
    sf_requirements = req_lines.splitlines()

with open(os.path.join(_here, 'requirements-bigquery.txt'), encoding='utf-8') as f:
    req_lines = f.read()
    bq_requirements = req_lines.splitlines()

setup(
    name='rasgoql',
    version=__version__,
    description=('Alpha version of rasgoQL open-source package.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rasgo Intelligence',
    author_email='patrick@rasgoml.com',
    project_urls={
        'Documentation': 'https://docs.rasgoql.com',
        'Source': 'https://github.com/rasgointelligence/RasgoQL',
        'Rasgo': 'https://www.rasgoml.com/',
    },
    license='GNU Affero General Public License v3 or later (AGPLv3+)',
    packages=[
        'rasgoql',
        'rasgoql.data',
        'rasgoql.primitives',
        'rasgoql.utils'
        ],
    install_requires=requirements,
    extras_require={
        "snowflake":  sf_requirements,
        "bigquery": bq_requirements,
    },
    include_package_data=True,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Database',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Code Generators'
        ]
)



"""
DataWarehouse credential functions
"""
import os
from pathlib import Path
import yaml

import dotenv


def load_env(
        filepath: Path = None
    ):
    """
    Loads a .env file, allowing contents to be callable by os.getenv()
    """
    if not filepath:
        filepath = os.getcwd()
    if not filepath.endswith('.env'):
        filepath = os.path.join(filepath, '.env')
    if not os.path.exists(filepath):
        raise FileNotFoundError(f'File {filepath} does not exist')
    dotenv.load_dotenv(filepath)

def load_yml(
        filepath: Path
    ):
    """
    Loads a yml file and returns the contents
    """
    with open(filepath) as yml_file:
        contents = yaml.load(yml_file, Loader=yaml.Loader)
    return contents

def read_file(
        filepath: Path = None
    ) -> str:
    """
    Reads the contents of a file
    """
    if not os.path.exists(filepath):
        raise FileNotFoundError(f'File {filepath} does not exist')
    with open(filepath, "r") as _file:
        contents = _file.read()
    return contents

def save_env(
        creds: dict,
        filepath: Path = None,
        overwrite: bool = False
    ):
    """
    Writes creds to a .env file
    """
    if not filepath:
        filepath = os.getcwd()
    if not filepath.endswith('.env'):
        filepath = os.path.join(filepath, '.env')
    if os.path.exists(filepath):
        if not overwrite:
            raise FileExistsError(
                f'File {filepath} already exist, pass in overwrite=True to overwrite it. '
                'Warning: this will overwrite all existing values in this .env file.')
    else:
        f = open(filepath, "x")
        f.close()
    for k, v in creds.items():
        dotenv.set_key(filepath, k, v)
    return filepath

def save_yml(
        creds: str,
        filepath: Path = None,
        overwrite: bool = True
    ):
    """
    Write creds to a .yml file
    """
    if filepath is None:
        filepath = os.getcwd()

    if filepath.endswith('/'):
        filepath = os.path.join(filepath, 'credentials.yml')

    if filepath.split(".")[-1] not in ['yaml', 'yml']:
        filepath = os.path.join(filepath, ".yml")

    if os.path.exists(filepath) and not overwrite:
        raise FileExistsError(
            f'{filepath} already exists. If you wish to overwrite it, '
            'pass overwrite=True and run this function again.')
    with open(filepath, "w") as _yaml:
        yaml.dump(data=creds, Dumper=yaml.SafeDumper, stream=_yaml)
    return filepath
